// This may look like C code, but it is really -*- C++ -*-
//
// history:
// - mComplex added on Apr99 for large systems to save memory 
// ToDo:
// - add fix point number x*2^(-15) as typedef, + put, get
// - class mComplex => typedef mComplex + put, get
//
/* typedef double zahl !!! */

#ifndef _Complex_h
#ifdef __GNUG__
/* #pragma interface */
#endif
#define _Complex_h 1


// #include <iostream.h>
#include <math.h>


class mComplex;		/* declaration Apr99 */

class Complex
{
#ifdef __ATT_complex__
public:
#else
protected:
#endif

  zahl           re;
  zahl           im;

public:

  zahl           real() const;
  zahl           imag() const;

                   Complex();
                   Complex(const  Complex& y);
                   Complex(const mComplex& y);
                   Complex(zahl r, zahl i=0);

                  ~Complex();

  Complex&         operator =  (const Complex& y);
  Complex&         operator =  (zahl y);               /* JS */

  Complex&         operator += (const Complex& y);
  Complex&         operator += (zahl y);
  Complex&         operator -= (const Complex& y);
  Complex&         operator -= (zahl y);
  Complex&         operator *= (const Complex& y);
  Complex&         operator *= (zahl y);

  Complex&         operator /= (const Complex& y); 
  Complex&         operator /= (zahl y); 

  void             error(const char* msg) const;
};

/* ====================================================================
   same again with float instead of double (templates are more risky?)
   ==================================================================== */
class mComplex
{
#ifdef __ATT_complex__
public:
#else
protected:
#endif

  mzahl           re;
  mzahl           im;

public:

  mzahl           mreal() const;
  mzahl           mimag() const;

                   mComplex();
                   mComplex(const mComplex& y);
                   mComplex(const  Complex& y);
                   mComplex(mzahl r, mzahl i=0);

                  ~mComplex();

  mComplex&        operator += (const mComplex& y);
  mComplex&        operator -= (const mComplex& y);
  mComplex&        operator *= (const mComplex& y);

  void             error(const char* msg) const;
};


inline mzahl  mComplex::mreal() const { return re; }
inline mzahl  mComplex::mimag() const { return im; }

// non-inline functions

#if 0
Complex   operator /  (const Complex& x, const Complex& y);
Complex   operator /  (const Complex& x, zahl y);
Complex   operator /  (zahl   x, const Complex& y);


Complex   cos(const Complex& x);
Complex   sin(const Complex& x);

Complex   cosh(const Complex& x);
Complex   sinh(const Complex& x);

Complex   exp(const Complex& x);
Complex   log(const Complex& x);

Complex   pow(const Complex& x, int p);
Complex   pow(const Complex& x, const Complex& p);
Complex   pow(const Complex& x, zahl y);
Complex   sqrt(const Complex& x);

istream&  operator >> (istream& s, Complex& x);
ostream&  operator << (ostream& s, const Complex& x);

// other functions defined as inlines

int  operator == (const Complex& x, const Complex& y);
int  operator == (const Complex& x, zahl y);
int  operator != (const Complex& x, const Complex& y);
int  operator != (const Complex& x, zahl y);

Complex  operator - (const Complex& x);
Complex  conj(const Complex& x);
Complex  operator + (const Complex& x, const Complex& y);
Complex  operator + (const Complex& x, zahl y);
Complex  operator + (zahl x, const Complex& y);
Complex  operator - (const Complex& x, const Complex& y);
Complex  operator - (const Complex& x, zahl y);
Complex  operator - (zahl x, const Complex& y);
Complex  operator * (const Complex& x, const Complex& y);
Complex  operator * (const Complex& x, zahl y);
Complex  operator * (zahl x, const Complex& y);

zahl  real(const Complex& x);
/* zahl  real(const zahl& x); */    /* JS */
zahl  imag(const Complex& x);
zahl  abs(const Complex& x);
zahl  fabs(const Complex& x); /* JS */
zahl  norm(const Complex& x);
zahl  arg(const Complex& x);

Complex polar(zahl r, zahl t = 0.0);

#endif
// inline members

inline zahl  Complex::real() const { return re; }
inline zahl  Complex::imag() const { return im; }

inline Complex::Complex() {}
inline Complex::Complex(const  Complex& y) :re(y.real()), im(y.imag()) {}
inline Complex::Complex(const mComplex& y) :re(y.mreal()), im(y.mimag()) {}
inline Complex::Complex(zahl r, zahl i) :re(r), im(i) {}

inline Complex::~Complex() {}

inline Complex&  Complex::operator =  (const Complex& y) 
{ re = y.real(); im = y.imag(); return *this;
} 

           /* JS */
inline Complex&  Complex::operator =  (zahl y) 
{ re = y; im = 0.0; return *this; } 


inline Complex&  Complex::operator += (const Complex& y)
{ re += y.real();  im += y.imag(); return *this;
}

inline Complex&  Complex::operator += (zahl y)
{ re += y; return *this;
}

inline Complex&  Complex::operator -= (const Complex& y)
{ re -= y.real();  im -= y.imag(); return *this;
}

inline Complex&  Complex::operator -= (zahl y)
{ re -= y; return *this;
}

inline Complex&  Complex::operator *= (const Complex& y)
{ zahl r = re * y.real() - im * y.imag();
  im = re * y.imag() + im * y.real(); 
  re = r; 
  return *this; 
}

inline Complex&  Complex::operator *= (zahl y)
{ re *=  y; im *=  y; return *this;}

inline Complex&  Complex::operator /= (const Complex& y) /* JS */
{ zahl r=(re*y.real()+im*y.imag())/(y.real()*y.real()+y.imag()*y.imag());
  im = (-re*y.imag()+im*y.real())/(y.real()*y.real()+y.imag()*y.imag());
  re = r; 
  return *this; 
}

inline Complex&  Complex::operator /= (zahl y)         /* JS */
{ re /=  y; im /=  y; return *this;}

//  functions

inline int  operator == (const Complex& x, const Complex& y)
{ return x.real() == y.real() && x.imag() == y.imag();}

inline int  operator == (const Complex& x, zahl y)
{ return x.imag() == 0.0 && x.real() == y;}

inline int  operator != (const Complex& x, const Complex& y)
{ return x.real() != y.real() || x.imag() != y.imag();}

inline int  operator != (const Complex& x, zahl y)
{ return x.imag() != 0.0 || x.real() != y;}

inline Complex  operator - (const Complex& x)
{ return Complex(-x.real(), -x.imag());}

inline Complex  conj(const Complex& x)
{ return Complex(x.real(), -x.imag());}

inline Complex  operator + (const Complex& x, const Complex& y)
{ return Complex(x.real() + y.real(), x.imag() + y.imag());}

inline Complex  operator + (const Complex& x, zahl y)
{ return Complex(x.real() + y, x.imag());}

inline Complex  operator + (zahl x, const Complex& y)
{ return Complex(x + y.real(), y.imag());}

inline Complex  operator - (const Complex& x, const Complex& y)
{ return Complex(x.real() - y.real(), x.imag() - y.imag());}

inline Complex  operator - (const Complex& x, zahl y)
{ return Complex(x.real() - y, x.imag());}

inline Complex  operator - (zahl x, const Complex& y)
{ return Complex(x - y.real(), -y.imag());}

inline Complex  operator * (const Complex& x, const Complex& y)
{ return Complex(x.real() * y.real() - x.imag() * y.imag(),
                 x.real() * y.imag() + x.imag() * y.real());}

inline Complex  operator * (const Complex& x, zahl y)
{  return Complex(x.real() * y, x.imag() * y);}

inline Complex  operator * (zahl x, const Complex& y)
{ return Complex(x * y.real(), x * y.imag());}

inline Complex  operator / (const Complex& x, const Complex& y) /* JS */
{ return Complex(
  ( x.real()*y.real()+x.imag()*y.imag())/(y.real()*y.real()+y.imag()*y.imag()),
  (-x.real()*y.imag()+x.imag()*y.real())/(y.real()*y.real()+y.imag()*y.imag()));}

inline Complex  operator / (const Complex& x, zahl y) /* JS */
{  return Complex(x.real() / y, x.imag() / y);}

inline Complex  operator / (zahl x, const Complex& y) /* JS */
{ return Complex(
   x*y.real()/(y.real()*y.real()+y.imag()*y.imag()),
  -x*y.imag()/(y.real()*y.real()+y.imag()*y.imag()));}

// inline zahl  real(const zahl& x) /* JS */
// { return x;}

inline zahl  real(const Complex& x)
{ return x.real();}

inline zahl  imag(const Complex& x)
{ return x.imag();}

inline zahl  abs(const Complex& x)
{ return hypot(x.real(), x.imag());}

inline zahl  fabs(const Complex& x) /* JS zeitintensiv! */
{ return hypot(x.real(), x.imag());}

inline zahl  norm(const Complex& x)
{ return (x.real() * x.real() + x.imag() * x.imag());}

inline zahl  arg(const Complex& x)
{ return atan2(x.imag(), x.real());}

inline Complex  polar(zahl r, zahl t)
{ return Complex(r * cos(t), r * sin(t));}

inline Complex  sqrt(const Complex& x)  /* JS */
{ return Complex(sqrt(abs(x))*cos(arg(x)/2.),
                 sqrt(abs(x))*sin(arg(x)/2.));}

/* ====================================================================
   same again with float instead of double (templates are more risky?)
   ==================================================================== */

// inline members


inline mComplex::mComplex() {}
// following line was out-commented until v1.7.8b, why? error if VecType=4
inline mComplex::mComplex(const mComplex& y) :re(y.mreal()), im(y.mimag()) {}
inline mComplex::mComplex(const  Complex& y) :re(y.real()), im(y.imag()) {}
inline mComplex::mComplex(mzahl r, mzahl i) :re(r), im(i) {}

inline mComplex::~mComplex() {}

inline mComplex&  mComplex::operator += (const mComplex& y)
{ re += y.mreal();  im += y.mimag(); return *this;
}
inline mComplex&  mComplex::operator -= (const mComplex& y)
{ re -= y.mreal();  im -= y.mimag(); return *this;
}

inline mComplex&  mComplex::operator *= (const mComplex& y)
{ mzahl r = re * y.mreal() - im * y.mimag();
  im = re * y.mimag() + im * y.mreal(); 
  re = r; 
  return *this; 
}

#endif
