#ifndef _DEFS_H_
#define _DEFS_H_
#define MaxMem		6000000l

#include "config.h"
#include "error.h"
#include "spinsdef.h"

#ifndef MAX_ITR		/* can be set now in config.h, Apr08 */
#define MAX_ITR 1000	/* maximum number of iterations, old=150 */
#endif

#if defined(__STRICT_ANSI__) || !(defined(__cplusplus) || defined(__GNUC__) || defined(__CPP__))
#define inline
#endif

/* please define VecType in config.h !!! Mar99 JS */
#ifndef VecType
#if defined(__cplusplus)
#define VecType      5     /* complex double */
#else
#define VecType      1	/* real double */
#endif
/*
    0=float,1=double,2=long dbl,4=complex-flt (c++),5=complex-dbl (c++) Apr99
    8=16bit dont_use!_test_phase
 */
#endif

/*  plan: use short only for large vectors (less memory)!

         VecType:           0          1        2           !4          
 typical | typedef  |  short-  | normal- | long-  |   real-version | z.Zeit
 --------+----------+----------+---------+--------+----------------+--------
 array     cplx_s     sComplex    Complex  Complex   real_s          mcplx
 sums      cplx_l     Complex     Complex  lComplex  real_l          cplx
 array     real_s     float       double   double    (always real)   mzahl
 sums      real_l     double      double   ldouble   (always real)   zahl

 cplx_x is identical to real_x in real mode (only real EVs)

*/

#if VecType & 4  /* non-double types only in c++, for better readability?! */  
#ifndef __cplusplus
#   error "use c++ !!!"
#endif
#endif

#if   (VecType & (3+8)) == 8          /* __GNUC__ 2.5.8 */
/* ====================================================================
   16-bit real number (non integer) for large arrays x*2^(-15)
   used via define mVecType, mVecType_get(), mVecType_put(), 
   ==================================================================== */
typedef short float16;
#include <math.h>
static double float16_base   = 1.0/pow(2,sizeof(float16)*8-2); /* slow! */
static double float16_1_base =     pow(2,sizeof(float16)*8-2); /* 1/base */
#include <stdio.h>
static inline double  float16_f2d(float16 x){  /* -> double */
  /*  printf(" f2d(%d*%g)=%g\n",(int)x,float16_base,x*float16_base); */
  return x*float16_base;
}
static inline float16 float16_d2f(double  x){  /* double -> */
  double y=x*float16_1_base+((x>=0)?0.5:-0.5);
  /* printf(" d2f(%g)=%d\n",x,(int)y); */
  if ((long)((float16)y)!=(long)y) ERR(300);
  return (float16)y;
}
typedef double zahl;  
typedef float16 mzahl;
#elif (VecType & 3) == 2
typedef long double zahl;  
typedef long double mzahl;  
#elif (VecType & 3) == 0
typedef double zahl;    
typedef float mzahl;    
#else
typedef double zahl;    
typedef double mzahl;    
#endif

#define Z(x)   ((zahl)(x))
#if VecType & 4
#define IFcplx(x) x       /* operationen, nur bei COMPLEX */
#include "complex.h"
#if   (VecType & 3) == 0          /* __GNUC__ 2.5.8 */
typedef class Complex cplx; 
typedef class mComplex mcplx; 
#else
typedef class Complex cplx; 
typedef class Complex mcplx; 
#endif
#define C(x,y)  Complex((zahl)x,(zahl)y)
#define D(x)   ((double)real(x))
#define Re(x)   real(x)
#define Im(x)   imag(x)
#define Conj(x) conj(x)
#define Norm(x) norm(x)   /* r*r+i*i */
#else  /* ! ( VecType & 4 ) */
#define IFcplx(x)
typedef zahl cplx;
typedef mzahl mcplx;
#define C(x,y)  ((zahl)(x))
#define D(x)   ((double)(x))
#define Re(x)   (x)
#define Im(x)   0
#define Conj(x) (x)
#define Norm(x) norm(x)   /* not needed!? */
/* #define Norm(x) fsqr(x) */
#endif

#if VecType & 8
#define M_d2f(x) float16_d2f(x)
#define M_f2d(x) float16_f2d(x)
#else
#define M_d2f(x) ((mcplx)x)
#define M_f2d(x) ((cplx)x)
#endif

#ifdef S_llong				/* Dec99 JS */
#if    S_llong>S_long			/* Jan00 JS */
 typedef unsigned long long ULLong;
#else
 typedef unsigned long ULLong;
#endif
#else
 typedef unsigned long ULLong;
#endif

typedef unsigned long ULong;  /* already defined? -> where? Warnings */

#define Dsize sizeof(mzahl)
#define Lsize sizeof(long)
#define Csize sizeof(mcplx)

/* using "\" in __TURBOC__ produces wrong line numbers!	*/
#define FF {}     	/* fflush(stdout) -> setvbuff(stdout,...)	*/

#ifdef sun
#undef sun                      /* no problems with 'SUN's (-Dsun) */
#endif

#include "array_m.h"		/* needs definition of mcplx */

/* SLES9 forgets min max if array_m.h follows, cplx=dbl */
#define max(a,b) (((a) > (b)) ? (a) : (b))
#define min(a,b) (((a) < (b)) ? (a) : (b))

#endif /* ifndef _DEFS_H_ */
